@@ -18,13 +18,22 @@
 
 #include "spellcheck.h"
 
+#include <QGuiApplication>
 #include <QClipboard>
 #include <QDebug>
 #include <QLocale>
 #include <QSet>
+#include <QIcon>
+#include <QMimeData>
 
 #include <KLocalizedString>
 
+namespace {
+namespace ActionIds {
+inline QString copyToClipboard() { return QStringLiteral("copyToClipboard"); }
+}
+}
+
 SpellCheckRunner::SpellCheckRunner(QObject* parent, const QVariantList &args)
     : Plasma::AbstractRunner(parent, args)
 {
@@ -32,6 +41,8 @@ SpellCheckRunner::SpellCheckRunner(QObject* parent, const QVariantList &args)
     setObjectName(QLatin1String( "Spell Checker" ));
     setIgnoredTypes(Plasma::RunnerContext::FileSystem | Plasma::RunnerContext::NetworkLocation);
     setSpeed(AbstractRunner::SlowSpeed);
+
+    addAction(ActionIds::copyToClipboard(), QIcon::fromTheme(QStringLiteral("edit-copy")), i18nc("@action", "Copy to Clipboard"));
 }
 
 SpellCheckRunner::~SpellCheckRunner()
@@ -51,12 +62,15 @@ void SpellCheckRunner::init()
 void SpellCheckRunner::loaddata()
 {
     //Load the default speller, with the default language
-    if (!m_spellers.contains("")) {
-        m_spellers[""] = QSharedPointer<Sonnet::Speller> (new Sonnet::Speller(""));
+    auto defaultSpellerIt = m_spellers.find(QString());
+    if (defaultSpellerIt == m_spellers.end()) {
+        defaultSpellerIt = m_spellers.insert(QString(), QSharedPointer<Sonnet::Speller> (new Sonnet::Speller(QString())));
     }
+    auto& defaultSpeller = defaultSpellerIt.value();
+
     //store all language names, makes it possible to type "spell german TERM" if english locale is set
     //Need to construct a map between natual language names and names the spell-check recognises.
-    const QStringList avail = m_spellers[""]->availableLanguages();
+    const QStringList avail = defaultSpeller->availableLanguages();
     //We need to filter the available languages so that we associate the natural language
     //name (eg. 'german') with one sub-code.
     QSet<QString> families;
@@ -70,17 +84,18 @@ void SpellCheckRunner::loaddata()
         QString code;
         //If we only have one code, use it.
         //If a string is the default language, use it
-        if (family.contains(m_spellers[""]->language())) {
-            code = m_spellers[""]->language();
+        if (family.contains(defaultSpeller->language())) {
+            code = defaultSpeller->language();
         } else if (fcode == QLatin1String("en")) {
             //If the family is english, default to en_US.
-            if (family.contains("en_US")) {
-                code = QLatin1String("en_US");
+            const auto enUS = QStringLiteral("en_US");
+            if (family.contains(enUS)) {
+                code = enUS;
             }
-        } else if (family.contains(fcode+QLatin1String("_")+fcode.toUpper())) {
+        } else if (family.contains(fcode+QLatin1Char('_')+fcode.toUpper())) {
             //If we have a speller of the form xx_XX, try that.
             //This gets us most European languages with more than one spelling.
-            code =  fcode+QLatin1String("_")+fcode.toUpper();
+            code =  fcode+QLatin1Char('_')+fcode.toUpper();
         } else {
             //Otherwise, pick the first value as it is highest priority.
             code = family.first();
@@ -127,8 +142,9 @@ void SpellCheckRunner::reloadConfiguration()
  * Return the empty string if we can't match a language. */
 QString SpellCheckRunner::findlang(const QStringList& terms)
 {
+    auto defaultSpeller = m_spellers[QString()];
     //If first term is a language code (like en_GB), set it as the spell-check language
-    if (terms.count() >= 1 && m_spellers[""]->availableLanguages().contains(terms[0])) {
+    if (terms.count() >= 1 && defaultSpeller->availableLanguages().contains(terms[0])) {
         return terms[0];
     }
     //If we have two terms and the first is a language name (eg 'french'),
@@ -152,7 +168,7 @@ QString SpellCheckRunner::findlang(const QStringList& terms)
 
         if (!code.isEmpty()) {
             //We found a valid language! Check still available
-            const QStringList avail = m_spellers[""]->availableLanguages();
+            const QStringList avail = defaultSpeller->availableLanguages();
             //Does the spell-checker like it?
             if (avail.contains(code)) {
                 return code;
@@ -160,7 +176,7 @@ QString SpellCheckRunner::findlang(const QStringList& terms)
         }
         //FIXME: Support things like 'british english' or 'canadian french'
     }
-    return QLatin1String("");
+    return QString();
 }
 
 void SpellCheckRunner::match(Plasma::RunnerContext &context)
@@ -181,10 +197,10 @@ void SpellCheckRunner::match(Plasma::RunnerContext &context)
     }
 
     //Pointer to speller object with our chosen language
-    QSharedPointer<Sonnet::Speller> speller = m_spellers[""];
+    QSharedPointer<Sonnet::Speller> speller = m_spellers[QString()];
 
     if (speller->isValid()) {
-        QStringList terms = query.split(' ', QString::SkipEmptyParts);
+        QStringList terms = query.split(QLatin1Char(' '), QString::SkipEmptyParts);
         QString lang = findlang(terms);
         //If we found a language, create a new speller object using it.
         if (!lang.isEmpty()) {
@@ -208,36 +224,60 @@ void SpellCheckRunner::match(Plasma::RunnerContext &context)
         return;
     }
 
-    Plasma::QueryMatch match(this);
-    match.setType(Plasma::QueryMatch::InformationalMatch);
-
     if (speller->isValid()) {
         QStringList suggestions;
         const bool correct = speller->checkAndSuggest(query,suggestions);
         if (correct) {
-            match.setIcon(QIcon::fromTheme(QLatin1String( "checkbox" )));
-            match.setText(i18n("Correct")+QLatin1String(": ")+query);
+            Plasma::QueryMatch match(this);
+            match.setType(Plasma::QueryMatch::InformationalMatch);
+            match.setIconName(QStringLiteral("checkbox"));
+            match.setText(query);
+            match.setSubtext(i18nc("Term is spelled correctly", "Correct"));
+            match.setData(query);
+            context.addMatch(match);
         } else {
-            match.setIcon(QIcon::fromTheme(QLatin1String( "edit-delete" )));
-            const QString recommended = i18n("Suggested words: %1", suggestions.join(i18nc("separator for a list of words", ", ")));
-            //TODO: try setting a text and a subtext, with the subtext being the suggestions
-            match.setText(recommended);
-            match.setData(suggestions);
+            for (const auto& suggestion : qAsConst(suggestions)) {
+                Plasma::QueryMatch match(this);
+                match.setType(Plasma::QueryMatch::InformationalMatch);
+                match.setIconName(QStringLiteral("edit-rename"));
+                match.setText(suggestion);
+                match.setSubtext(i18n("Suggested term"));
+                match.setData(suggestion);
+                context.addMatch(match);
+            }
         }
     } else {
-        match.setIcon(QIcon::fromTheme(QLatin1String("task-attention")));
+        Plasma::QueryMatch match(this);
+        match.setType(Plasma::QueryMatch::InformationalMatch);
+        match.setIconName(QStringLiteral("task-attention"));
         match.setText(i18n("Could not find a dictionary."));
+        context.addMatch(match);
     }
-
-    context.addMatch(match);
 }
 
 void SpellCheckRunner::run(const Plasma::RunnerContext &context, const Plasma::QueryMatch &match)
 {
     Q_UNUSED(context)
-    //Copy words to clipboard
-    // FIXME:
-    // kapp->clipboard()->setText(match.data().toString());
+    if (match.selectedAction() == action(ActionIds::copyToClipboard())) {
+        //Copy words to clipboard
+        const QString text = match.data().toString();
+        QGuiApplication::clipboard()->setText(text);
+    }
+}
+
+QList<QAction *> SpellCheckRunner::actionsForMatch(const Plasma::QueryMatch &match)
+{
+    Q_UNUSED(match)
+
+    return {action(ActionIds::copyToClipboard())};
+}
+
+QMimeData * SpellCheckRunner::mimeDataForMatch(const Plasma::QueryMatch &match)
+{
+    QMimeData *result = new QMimeData();
+    const QString text = match.data().toString();
+    result->setText(text);
+    return result;
 }
 
 K_EXPORT_PLASMA_RUNNER(krunner_spellcheck, SpellCheckRunner)
